#-*- coding:utf-8 -*-
""" 
$Id$
"""
import unittest
import transaction
from datetime import date

from zope.interface import implements
from zope.interface.verify import verifyObject, verifyClass
from zope.app import zapi
from zope.rdb.interfaces import IZopeDatabaseAdapter

from zsqlmap import SQLObject
from zsqlmap.cols import *

from zsqlmap import DefaultError, ValidateError
from zsqlmap.interfaces import IORMManager, ISQLObject, IResult


class Account(SQLObject):
    title = StringCol()
    amount = IntCol(default=0)

class BaseTestORM(unittest.TestCase):

    def getDA(self):
        raise NotImplementedError('must be implemented')

    def getORMManager(self, da_name):
        raise NotImplementedError('must be implemented')

    def getCount(self, cls):
        sql = "SELECT COUNT(1) FROM %s" % cls.table
        cur = self.orm.connection.cursor()
        cur.execute(sql)
        row = cur.fetchone()
        return int(row[0])
        
    def setUp(self):
        da = self.getDA()
        orm = self.getORMManager('myda')
        self.da = da
        self.assert_( verifyObject(IZopeDatabaseAdapter,da))
                
        gsm = zapi.getGlobalSiteManager()        
        gsm.registerUtility(da, IZopeDatabaseAdapter,name='myda')
        gsm.registerUtility(orm, IORMManager)

        self.orm = orm
        self.orm.configure_for_request()

    def test_orm(self):
        orm = zapi.getUtility(IORMManager)
        self.assert_(verifyObject(IORMManager,orm))

        sql = ';'.join(orm.createTableSQL(Account))
        self.assert_( 'CREATE TABLE' in sql)
        self.assert_( 'title' in sql)
        self.assert_( 'amount' in sql)
                
        # создаём таблицу
        orm.createTable(Account)
        transaction.commit()
        
        prev_count = self.getCount(Account)
        # создаём объект
        o = Account(title='title1', amount=124)
        
        self.assert_(ISQLObject.providedBy(o))

        self.assert_(  prev_count + 1, self.getCount(Account) )
        self.assert_(o.id)                
        self.assertEqual(o.title, 'title1')
        self.assertEqual(o.amount, 124)

        self.assertEqual(o.table, 'account')

        columns = map(lambda x: x.name,o.columns)
        columns.sort()
        self.assertEqual( columns, ['amount', 'title'])
        
        self.assertEqual(o.fields['title'], 'title1')
        self.assertEqual(o.fields['amount'], 124)
        self.assertFalse(o.fields.has_key('id') )

        # singlenton in thread
        prev_count = self.getCount(Account)
        o_id = o.id
        o2 = Account.get(o_id)
        self.assertEqual(  prev_count, self.getCount(Account) )
        self.assert_(o2.id)
        self.assertEqual( id(o2), id(o) )      
       
        # transaction
        prev_count = self.getCount(Account)
        o.title = 'Bob' 
        transaction.commit()

        o = Account.get(o_id)
        self.assertEqual(o.id, o_id)
        self.assertEqual(o.title, 'Bob')
        self.assertEqual(o.amount, 124)
        
        o.title = 'Fred' 
        transaction.abort()        
        o2 = Account.get(o_id) # after abort, transaction cache must be clear
        self.assertNotEqual(id(o), id(o2)) 
        self.assertEqual(o2.id, o_id)
        self.assertEqual(o2.title, 'Bob')
        self.assertEqual(o2.amount, 124)

        self.assertEqual(  prev_count, self.getCount(Account) )
                
        # select
        res = Account.select(" id > 0 ")
        self.assert_(verifyObject(IResult, res))
        
        self.assertEqual( len(res), self.getCount(Account) )

        r = res[0] # [index]
        self.assert_(isinstance(r,Account))
        
        r = res[10:20] # [i:j]
        self.assert_(verifyObject(IResult, r))
        
        res = Account.select(" id > 0 AND 2>3 ")
        self.assertEqual( len(res), 0 )

    
        # selectBy
        res = Account.selectBy(title='Bob')
        self.assert_(verifyObject(IResult, res))

        self.assert_( len(res) > 0)
        for a in res:
            self.assertEqual(a.title, 'Bob')
        
    def test_result(self):
        res = Account.select()
        self.assert_(verifyObject(IResult, res))

        r = res[20:30]
        self.assert_(verifyObject(IResult, res))
        self.assertEqual(r.start, 20)
        self.assertEqual(r.end, 30)
        sql = r.finalPrepareQuery('')
        if r.native_slice:
            self.assert_('20' in sql)

        # aggregate
        res = Account.select()
        agr = res.aggregate('SUM(1)')
        self.assertEqual(agr, len(res))
        
    
    def test_foreign_key(self):
        class Item(SQLObject):            
            amount = IntCol(default=1)         
            oinfo = ForeignKey('OrderInfo') 
            zinfo = ForeignKey('OrderInfo')
            
        class OrderInfo(SQLObject):
            title = StringCol()

        orm = zapi.getUtility(IORMManager)
        
        orm.createTable(Item)
        orm.createTable(OrderInfo)
        transaction.commit() 
        
        o = OrderInfo(title='order1')
        o2 = OrderInfo(title='order2')        
        item = Item(amount=23, oinfo=o, zinfo=o2)
        
        self.assertEqual(o.id, item.oinfo.id)
        
        o_id = o.id
        o2_id = o2.id
        i_id = item.id
        #transaction.commit()
                
        i = Item.get(i_id)
        self.assertEqual(i.oinfo.id, o_id)
        self.assertEqual(i.zinfo.id, o2_id)
    

    def test_one_many(self):
        class Box(SQLObject):
            title = StringCol()
            items = MultipleJoin('BoxItem') 
            items_orders = MultipleJoin('BoxItem', orderBy='title')
            
        class BoxItem(SQLObject):
            title = StringCol()
            box = ForeignKey('Box')


        orm = zapi.getUtility(IORMManager)        
        orm.createTable(Box)
        orm.createTable(BoxItem)
        transaction.commit() 
        
        box1 = Box(title='box1')
        box2 = Box(title='box2')
        self.assertEqual( len(list(box1.items)), 0 )
        self.assertEqual( len(list(box2.items)), 0 )

        bi = BoxItem(title='bi1', box=box1)
        self.assertEqual( len(list(box1.items)), 1 )
        self.assertEqual( len(list(box2.items)), 0 )
        

    def test_cache(self):
        class MyObject(SQLObject):
            title = StringCol()

        self.orm.createTable(MyObject)
        transaction.commit()
        
        o = MyObject(title='title1')
        o_id = o.id

        o2 = MyObject.get(o_id)
        self.assertEqual(id(o), id(o2))
        o.title = 'title2'
        
        transaction.commit()
        
        o2 = MyObject.get(o_id)
        self.assertNotEqual(id(o), id(o2))
        self.assertEqual( o2.title, 'title2')
        
        
    def test_delete(self):
        class DeleteTest(SQLObject):
            title = StringCol()

        self.orm.createTable(DeleteTest)
        transaction.commit()
        
        o = DeleteTest(title='title1')
        o_id = o.id

        o.delete()
        transaction.commit()
        
        o2 = DeleteTest.get(o_id)
        self.assertEqual(o2, None)
        
        

         
